/*
 *  io.c : Enthaelt die Routinen fuer die Sockets und Logfiles
 */

/*-----------------------------------------------------------------------*/

#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include "config.h"
#include "io.h"
#include "struct.h"

#define BUFSIZE 2048
char mtphost[BUFSIZE];
int mtpport;

/*-----------------------------------------------------------------------*/

extern struct mtpheader *header;
extern struct mtpmessage *mtpm;
extern char *filebuffer;

/*-----------------------------------------------------------------------*/

int create_socket(port)
int port;

/*
 * create_socket() oeffnet eine Socket (s), nagelt sie auf der Adressstruktur
 * (sin) fest und geht wartet auf eine Verbindung (listen). Kommt diese
 * zustande, wird ueber accept() ein Level1-Filehandler zurueckgegeben.
 * Diese Routine ist das mieseste Stueck Patchwork, was ich je gesehen habe,
 * aber sie funktioniert (und sichert sich damit ihre Daseinsberechtigung).
 */

{
        static int s=0;
	static int g,len;
	static struct sockaddr_in sin,from;

	len = sizeof (&from);
	if (!s)
        {
		memset((char *)&sin,0,sizeof(sin));
		sin.sin_port = htons(port);
		sin.sin_addr.s_addr= htonl (INADDR_ANY);
		sin.sin_family = AF_INET;

		if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
		{
			return(0);
		}
		if (bind (s, (caddr_t)&sin,sizeof(sin)) < 0)
		{
			close(s);
			return(0);
		}

	}
	
	listen(s,5);
       	g=accept(s,&from,&len);
        return(g);
}

/*------------------------------------------------------------------------*/

int connect_socket(host,port)
char *host;
int port;

/*
 * connect_socket() versucht eine Verbindung zum Serverhost aufzustellen
 * (open_socket). Gelingt das, werden die angegebenen Daten uebertragen.
 */

{
int s;

	if (( s = open_socket(host,port)) == -1)
		return(0);

	return(s);
}  

/*------------------------------------------------------------------------*/

int open_socket(h_name,portnr)
char *h_name;
int portnr;

/*
 * open_socket() versucht eine Verbindung zum Serverhost herzustellen.
 * Gelingt dies, wird die Socketnummer zurueckgegeben, andernfalls -1.
 */

{
int s;
struct in_addr addr;
struct sockaddr_in sin;
struct hostent *host;

	if (!((*h_name >= '0') && (*h_name <= '9')))
	  {
	    host = gethostbyname(h_name);
	    if (host == NULL)
	      {
		printf("open_socket: Couldn't find host %s .\n",h_name);
		return 0;
	      }
	    h_name = (char *) inet_ntoa(host->h_addr_list[0]);
	  }

        addr.s_addr = inet_addr(h_name);
        if ((int)addr.s_addr == -1)
          {
	     printf("open_socket: Couldn't find host %s .\n",h_name);
	     return 0;
	  }

	memset((char *)&sin,0,sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(portnr);
	memcpy((char *)&sin.sin_addr,&addr,sizeof(struct sockaddr_in));

	if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("open_socket: socket");
		return -1 ;
	}

	if(connect(s, (char *)&sin, sizeof(sin)) < 0) {
		perror("open_socket: connect");
		return -1 ;
	}

	return s ;
}

/*------------------------------------------------------------------------*/

int log_message (mtph)
struct mtpheader *mtph;

/*
 * log_message baut einen String folgender Form auf:
 * Tag Monat Datum hh:mm:ss <Wizard> Filename (# Bytes)
 * Diese Message wird an das MTPLOGFILE appended.
 */

{
FILE *fd;
long t;
char timestr[30],tmp[80],mode[10];

	time (&t);
	sprintf (timestr,"%s",ctime(&t));
	if (mtph->flags&RECEIVE)
		strcpy(mode,"recv");
	else
		strcpy(mode,"send");
	timestr[strlen(timestr)-6] = (char)0;  /* Das unnuetze Jahr aus dem String nehmen */
	sprintf (tmp,"%s <%s> %s: %s (%d Bytes)\n",timestr,mtph->player,mode,mtph->filename,mtph->filelength);

	if (!(fd = fopen (MTPLOGFILE,"a")))
	{
		printf ("Couldn't write to MTPLOGFILE ... check config.h for the correct path !\n");
		close_all();
		exit(0);
	}
	fprintf (fd,tmp);
	fclose (fd);
}

/*--------------------------------------------------------------------------*/

char *introduce(length)
int length;

/*
 * introduce() ist meine Standardinitialisierung fuer Puffer, Structs etc.
 * Es ist nichts als ein leicht geschoentes malloc() mit dem Zusatz, dass
 * der Speicher sicher geleert wird. Sieht auf den ersten Blick umstaendlich
 * aus, erspart aber VIEL Aerger mit Verfuegbarkeitsbereichen etc.
 */

{
char *pointer;

	pointer = (char *) malloc (length);
	memset (pointer,0,length);
	return (pointer);
}

/*--------------------------------------------------------------------------*/

void dismiss(pointer)
char *pointer;

/*
 * dismiss() gibt die mit introduce() initialisierten Variablen wieder frei
 */
{
	free (pointer);
}

/*-------------------------------------------------------------------------*/

int safe_write (sh,buffer,length)
int sh;
char *buffer;
long length;

{
char *point;
int count,bytes,cy;

	count=length;
	point=buffer;
	cy=0;

	while (count>0)
	{
		bytes = write (sh,point,count);
		if (bytes<1)
			cy++;
		else
			cy=0;

		point+=bytes;
		count-=bytes;

		if (cy>10)
			return 0;
	}
	return 1;
}

/*------------------------------------------------------------------------*/

int safe_read (sh,buffer,length)
int sh;
char *buffer;
long length;

{
char *point;
int count,bytes,cy;

	count=length;
	point=buffer;
	cy=0;

	while (count>0)
	{
		bytes = read (sh,point,count);
		if (bytes<1)
			cy++;
		else
			cy=0;

		point+=bytes;
		count-=bytes;

		if (cy>10)
			return 0;
	}
	return 1;
}

/*-------------------------------------------------------------------------*/

int read_message (sh,m)
int sh;
struct mtpmessage *m;

{
	if (!(safe_read(sh,m,sizeof(struct mtpmessage))))
		return 0;

	printf (m->message);
	if (m->quit==1)
		return 0;

	return 1;
}

/*-----------------------------------------------------------------------*/

int write_message(sh,mtpm,message,quit)
int sh,quit;
struct mtpmessage *mtpm;
char *message;

{
int bytes;

	strcpy (mtpm->message,message);
	mtpm->quit = quit;
	bytes = safe_write (sh,mtpm,sizeof (struct mtpmessage));

	if (bytes<1)
		return 0;
	else
		return 1;
}

/*-----------------------------------------------------------------------*/

int mtp_open(path,mtph)
char *path;
struct mtpheader *mtph;

{
char tmp[200];
int fh;

	fh = open (path,O_WRONLY|O_CREAT|O_TRUNC);

	if (fh>0)
	{
		sprintf(tmp,"chmod 0600 %s",path);
		system(tmp);
	}

	return (fh);
}

/*-----------------------------------------------------------------------*/

char *build_path (mtph)
struct mtpheader *mtph;

{
static char path[100];

	if (mtph->flags&VERSION_TWO)
	{
		strcpy (path,MUDPATH);
		if (*mtph->filename!='/')
		{
			cat_path (path,"players");
			cat_path (path,mtph->player);
		}
		cat_path (path,mtph->filename);
	}
	else
	{
		strcpy (path,MUDPATH);
		cat_path (path,"players");
		cat_path (path,mtph->player);
		if (*mtph->directory!=(char)0)
			cat_path (path,cut_path(mtph->directory));
		cat_path (path,cut_path(mtph->filename));
	}
	return(path);
}

/*-----------------------------------------------------------------------*/

char *cut_path (path)
char *path;

{
int i;
char *point;

	if (*path==(char)0) return (path);
	for (i=strlen(path);i>0;i--)
	{
		point= &path[i+1];
		switch(path[i])
		{
			case '/' :	return (point);
					break;

			case '~' :	return (point);
					break;

			case '.' :	if (path[i-1]=='.')
						return (point);
					break;
		}
	}
        while (*path == '/') path++;
	return (path);
}

/*-----------------------------------------------------------------------*/

void send_file (sh,path,mtph)
int sh;
char *path;
struct mtpheader *mtph;

{
int fh,bytes;
char tmp[200];

	fh = open (path,O_RDONLY,4);
	if (fh<1)
	{
		sprintf (tmp,"Reception error: File not found or accessible !\n");
		write_message (sh,mtpm,tmp,1);
		return;
	}

	bytes = read (fh,filebuffer,MAX_FILELENGTH);
	if (bytes>=MAX_FILELENGTH)
	{
		sprintf (tmp,"Reception error: File too large. Maximum filesize is %d Bytes !\n",MAX_FILELENGTH);
		write_message (sh,mtpm,tmp,1);
		return;
	}
	close(fh);

	header->filelength=bytes;

	sprintf (tmp,"Connected to MTPserver V1.4 at %s.\nReception of file %s (%d Bytes) ...",mtphost,mtph->filename,mtph->filelength);
	if (!(write_message (sh,mtpm,tmp))) return;

	if (!(safe_write (sh,(char *)header,sizeof(struct mtpheader))))
	{
		return;
	}
	if (!(safe_write (sh,filebuffer,header->filelength)))
	{
		return;
	}

	sprintf (tmp," complete.\n");
	write_message (sh,mtpm,tmp);
	log_message(mtph);
}

/*--------------------------------------------------------------------------*/

void get_file (file,mtph)
char *file;
struct mtpheader *mtph;

{
char *path,tmp[200];
int sh,fh;

	header->filelength=0;
	fh = access(file,F_OK);

	if (fh==0)
	{
		printf ("Error: Cannot write to a existing file ... aborted\n");
		close_all();
		exit(0);
	}

	if (!(sh=connect_socket(mtphost,mtpport)))
	{
		printf ("No connection to MTPserver. Transmission failed !\n");
		close_all();
		exit(0);
	}

	strcpy (tmp,"@");
	safe_write (sh,tmp,1);

	strcpy (header->filename,file);

#ifdef ULTRIX
	SWAP (header->flags);
#endif
	if (!(safe_write (sh,(char *)header,sizeof(struct mtpheader))))
	{
		printf ("\nCouldn't write MTPheader ... aborting\n");
		close_all();
		exit (0);
	}

	if(!(read_message(sh,mtpm)))
		return;

	if (!(safe_read (sh,mtph,sizeof(struct mtpheader))))
	{
		printf ("\nCouldn't read MTPheader ... aborting\n");
		close_all();
		exit (0);
	}

#ifdef ULTRIX
	SWAP (mtph->filelength);
#endif
	if (*mtph->filename=='/')
		path = cut_path (mtph->filename);
	else
		path = mtph->filename;

	fh = open (path,O_WRONLY|O_CREAT|O_EXCL);
	if (fh<1)
	{
		fh = open (cut_path(path),O_WRONLY|O_CREAT|O_EXCL);
		if (fh<1)
		{
			strcpy (tmp,"\nCouldn't write destination file (existing ?) ... Transmission aborted.\n");
			close_all();
			exit (0);
		}
		strcpy(path,cut_path(path));
	}

	if (!(safe_read(sh,filebuffer,mtph->filelength)))
	{
		printf (" terminated.\nError during file transfer ...\n");
		close_all();
		exit (0);
	}
	else
	{
		printf (" completed.\n");
	}

	safe_write (fh,filebuffer,mtph->filelength);

	sprintf (tmp,"chmod 0600 %s\n",path);
	system (tmp);
	if (fh) close (fh);
}

/*-------------------------------------------------------------------------*/

void print_header(mtph)
struct mtpheader *mtph;
{
	printf ("Name : %s\n",mtph->player);
	printf ("Passwort : %s\n",mtph->passwd);
	printf ("Dateiname : %s\n",mtph->filename);
	printf ("Dateilaenge : %d\n",mtph->filelength);
	printf ("Status : %d\n",mtph->flags);
}

/*-------------------------------------------------------------------------*/

void get_params(thisgame,host,port,wizard,password)
char *thisgame, *host, *wizard, *password;
int *port;
{
char *hdir;
char buf[BUFSIZE];
FILE *cfile;
char game[BUFSIZE];
struct stat stb;

	hdir = (char *) getenv("HOME");
	if (hdir == NULL)
		hdir = ".";
	sprintf(buf, "%s/.mtprc", hdir);
	if (!(cfile = fopen(buf,"r")))
	{
		if (!(cfile = fopen(".mtprc","r")))
		{
			printf ("Couldn't read .mtprc !! Adjust it and copy it to your Homedir.\n");
			exit (1);
		}
	}

	if (fstat(fileno(cfile), &stb) >= 0 && (stb.st_mode & 077) != 0)
	{
		printf("Error - .mtprc file not correct mode. First do chmod 700 ~/.mtprc\n");
		exit(1);
	}
  
  while (fgets(buf, BUFSIZE, cfile) != NULL)
    {
      if (sscanf(buf, "game %s host %s port %u wizard %s password %s",
		 game, host, port, wizard, password) ==5)
	{
	  if (strcmp(game,thisgame) == 0)
	    {
	      /* got it */
	      return;
	    }
	}
      else if (thisgame[0]==0)
	{
	  if (sscanf(buf, "default %s", thisgame) == 1);
	  else printf("Mangled line in .mtprc: %s", buf);
	}
    }
  printf("Game %s not found. Check your .mtprc !\n", thisgame);
  exit(1);
  
}  

/*------------------------------------------------------------------------*/

char *cat_path (path,add)
char *path,*add;

{
char *point;

	point = add;
	if (strlen(path)==0)
		return(strcpy(path,add));
	if ((path[strlen(path)-1]=='/')&&(add[0]=='/'))
		point++;
	if ((path[strlen(path)-1]!='/')&&(add[0]!='/'))
		strcat (path,"/");
	strcat (path,point);
	return (path);
}

/*-----------------------------------------------------------------------*/

/* illegal_path() returns
 *
 * 0 : Path is OK, access is granted
 * 1 : Path contains illegal characters (~ or ..)
 * 2 : Normal wizard tries to write out of his own directories
 * 3 : Normal wizard tries to read from a locked directory
 *
 */

int illegal_path (path,level)
char *path;
int level;

{
char *point,tmp[200];

	strcpy (tmp,path);
	point = tmp;
	while (*point!=(char)0)     /* illegal fiddling with ~ and .. ? */
	{
		if (point[0]=='~')
			return 1;
		if ((point[0]=='.')&&(point[1]=='.'))
			return 1;
		point++;
	}

	if (level>=ARCHWIZ_LEVEL)
		return 0;		/* Archwizards have full read/write-permission */

	if (*header->filename!='/')
		return 0;		/* Wizard accesses his own directories -> OK */
		
	if (strncmp(header->filename,"/open/",6)==0)
		return 0;		/* All wizards may read/write to /open */

	if ((header->flags&RECEIVE)==0)
		return 2;		/* Wizard tries to write in outer directories */

	point = &(tmp[strlen(tmp)]);
	do
	{
		point--;
	}
	while (*point!='/');
	*point=(char)0;			/* cut the path at the latest slash */
	cat_path (tmp,LOCKFILE);

	if (access(tmp,0)==-1)
		return 0;
	
	return 3;
}

/*-------------------------------------------------------------------------*/

int check_directory(path)
char *path;

{
DIR *direct;

	direct = opendir (path);
	if (direct)
		return 1;
	else
		return 0;
}
